import matplotlib.pyplot as plt

def subplots(nrows=1, ncols=1, sharex=False, sharey=False, squeeze=True, 
            subplot_kw=None, hemisphere='lower', projection='equal_area', 
            **fig_kw):
    """
    Identical to matplotlib.pyplot.subplots, except that this will default to 
    producing equal-area stereonet axes.

    This prevents constantly doing:

        >>> fig, ax = plt.subplot(subplot_kw=dict(projection='stereonet'))

    or

        >>> fig = plt.figure()
        >>> ax = fig.add_subplot(111, projection='stereonet')

    Using this function also avoids having ``mplstereonet`` continually appear 
    to be an unused import when one of the above methods are used.

    Parameters:
    -----------
        nrows : int
          Number of rows of the subplot grid.  Defaults to 1.

        ncols : int
          Number of columns of the subplot grid.  Defaults to 1.

        hemisphere : string
            Currently this has no effect. When upper hemisphere and dual 
            hemisphere plots are implemented, this will control which hemisphere
            is displayed.

        projection : string
            The projection for the axes. Defaults to 'equal_area'--an equal-area
            (a.k.a. "Schmidtt") stereonet. May also be 'equal_angle' for an 
            equal-angle (a.k.a. "Wulff") stereonet or any other valid matplotlib
            projection (e.g. 'polar' or 'rectilinear' for a "normal" axes).

        The following parameters are identical to matplotlib.pyplot.subplots:

        sharex : string or bool
          If *True*, the X axis will be shared amongst all subplots.  If
          *True* and you have multiple rows, the x tick labels on all but
          the last row of plots will have visible set to *False*
          If a string must be one of "row", "col", "all", or "none".
          "all" has the same effect as *True*, "none" has the same effect
          as *False*.
          If "row", each subplot row will share a X axis.
          If "col", each subplot column will share a X axis and the x tick
          labels on all but the last row will have visible set to *False*.

        sharey : string or bool
            If *True*, the Y axis will be shared amongst all subplots. If
            *True* and you have multiple columns, the y tick labels on all but
            the first column of plots will have visible set to *False*
            If a string must be one of "row", "col", "all", or "none".
            "all" has the same effect as *True*, "none" has the same effect
            as *False*.
            If "row", each subplot row will share a Y axis.
            If "col", each subplot column will share a Y axis and the y tick
            labels on all but the last row will have visible set to *False*.

        *squeeze* : bool
            If *True*, extra dimensions are squeezed out from the
            returned axis object:

            - if only one subplot is constructed (nrows=ncols=1), the
              resulting single Axis object is returned as a scalar.

            - for Nx1 or 1xN subplots, the returned object is a 1-d numpy
              object array of Axis objects are returned as numpy 1-d
              arrays.

            - for NxM subplots with N>1 and M>1 are returned as a 2d
              array.

           If *False*, no squeezing at all is done: the returned axis
            object is always a 2-d array contaning Axis instances, even if it
            ends up being 1x1.

        *subplot_kw* : dict
            Dict with keywords passed to the
            :meth:`~matplotlib.figure.Figure.add_subplot` call used to
            create each subplots.

        *fig_kw* : dict
            Dict with keywords passed to the :func:`figure` call.  Note that all
            keywords not recognized above will be automatically included here.

    Returns:
    --------

        fig, ax : tuple

            - *fig* is the :class:`matplotlib.figure.Figure` object

            - *ax* can be either a single axis object or an array of axis
               objects if more than one supblot was created.  The dimensions
               of the resulting array can be controlled with the squeeze
               keyword, see above.
    """
    if projection in ['equal_area', 'equal_angle']:
        projection += '_stereonet'
    if subplot_kw == None:
        subplot_kw = {}
    subplot_kw['projection'] = projection
    return plt.subplots(nrows, ncols, sharex=sharex, sharey=sharey, 
                        squeeze=squeeze, subplot_kw=subplot_kw, **fig_kw)




from stereonet_axes import StereonetAxes
from stereonet_math import pole, plane, line, rake, plunge_bearing2pole
from stereonet_math import geographic2pole, geographic2plunge_bearing
from stereonet_math import xyz2stereonet, stereonet2xyz, azimuth2rake
from stereonet_math import vector2plunge_bearing, vector2pole
from stereonet_math import plane_intersection, project_onto_plane
from contouring import density_grid
from utilities import parse_quadrant_measurement, parse_strike_dip, parse_rake
from utilities import parse_azimuth, parse_plunge_bearing
from convenience_functions import subplots


__all__ = ['StereonetAxes', 'pole', 'plane', 'line', 'rake', 
           'plunge_bearing2pole', 'geographic2pole', 'vector2plunge_bearing',
           'geographic2plunge_bearing', 'density_grid', 'plane_intersection',
           'xyz2stereonet', 'stereonet2xyz', 'vector2pole', 
           'project_onto_plane', 'azimuth2rake', 'parse_azimuth',
           'parse_quadrant_measurement', 'parse_strike_dip', 'parse_rake',
           'parse_plunge_bearing', 'subplots']



import numpy as np

import matplotlib as mpl
from matplotlib.transforms import Affine2D
from matplotlib.projections import register_projection, LambertAxes
from matplotlib.axes import Axes
from matplotlib.ticker import NullLocator

import stereonet_math
import contouring
import stereonet_transforms

class StereonetAxes(LambertAxes):
    """An axes representing a lower-hemisphere "schmitt" (a.k.a. equal area) 
    projection."""
    name = 'stereonet'
    RESOLUTION = 30
    _base_transform = stereonet_transforms.LambertTransform

    def __init__(self, *args, **kwargs):
        """Initialization is identical to a normal Axes object except for the
        following kwarg:

        Parameters:
        -----------
            rotation : number
                The rotation of the stereonet in degrees clockwise from North.

        All additional args and kwargs are identical to Axes.__init__
        """
        # There's also center_latitude and center_longitude, but I'm 
        # deliberately keeping these undocumented until they're fully supported.
        self.horizon = np.radians(90)
        self._rotation = -np.radians(kwargs.pop('rotation', 0))
        LambertAxes.__init__(self, *args, **kwargs)

    def _get_core_transform(self, resolution):
        """The projection for the stereonet as a matplotlib transform. This is
        primarily called by LambertAxes._set_lim_and_transforms."""
        return self._base_transform(self._center_longitude,
                                    self._center_latitude,
                                    resolution)

    def _get_affine_transform(self):
        """The affine portion of the base transform. This is called by 
        LambertAxes._set_lim_and_transforms."""
        transform = self._get_core_transform(self.RESOLUTION)
        # How big is the projected globe?
        xscale, _ = transform.transform_point((self.horizon, 0))  
        _, yscale = transform.transform_point((0, np.pi / 2.0))
        # Create an affine transform to stretch the projection from 0-1
        return Affine2D() \
            .rotate(np.radians(self.rotation)) \
            .scale(0.5 / xscale, 0.5 / yscale) \
            .translate(0.5, 0.5)

    def _set_lim_and_transforms(self):
        """Setup the key transforms for the axes.""" 
        # Most of the transforms are set up correctly by LambertAxes
        LambertAxes._set_lim_and_transforms(self)

        # Transform for latitude ticks. These are typically unused, but just
        # in case we need them...
        yaxis_stretch = Affine2D().scale(2 * self.horizon, 1.0)
        yaxis_stretch = yaxis_stretch.translate(-self.horizon, 0.0)

        # These are identical to LambertAxes._set_lim_and_transforms, but we 
        # need to update things to reflect the new "yaxis_stretch"
        yaxis_space = Affine2D().scale(1.0, 1.1)
        self._yaxis_transform = \
            yaxis_stretch + \
            self.transData
        yaxis_text_base = \
            yaxis_stretch + \
            self.transProjection + \
            (yaxis_space + \
             self.transAffine + \
             self.transAxes)
        self._yaxis_text1_transform = \
            yaxis_text_base + \
            Affine2D().translate(-8.0, 0.0)
        self._yaxis_text2_transform = \
            yaxis_text_base + \
            Affine2D().translate(8.0, 0.0)

    def set_position(self, pos, which='both'):
        """Identical to Axes.set_position (This docstring is overwritten)."""
        self._polar.set_position(pos, which)
        LambertAxes.set_position(self, pos, which)
    set_position.__doc__ = Axes.set_position.__doc__

    def set_rotation(self, rotation):
        """Set the rotation of the stereonet in degrees clockwise from North."""
        self._rotation = np.radians(rotation)
        self._polar.set_theta_offset(self._rotation + np.pi / 2.0)
        self.transData.invalidate()
        self.transAxes.invalidate()
        self._set_lim_and_transforms()
    def get_rotation(self):
        """The rotation of the stereonet in degrees clockwise from North."""
        return np.degrees(self._rotation)
    rotation = property(get_rotation, set_rotation)

    def cla(self):
        """Identical to Axes.cla (This docstring is overwritten)."""
        Axes.cla(self)

        # Set grid defaults...
        self.set_longitude_grid(10)
        self.set_latitude_grid(10)
        self.set_longitude_grid_ends(80)

        # Hide all ticks and tick labels for the "native" lon and lat axes
        self.xaxis.set_minor_locator(NullLocator())
        self.yaxis.set_minor_locator(NullLocator())
        self.xaxis.set_ticks_position('none')
        self.yaxis.set_ticks_position('none')
        self.xaxis.set_tick_params(label1On=False)
        self.yaxis.set_tick_params(label1On=False)

        # Set the grid on or off based on the rc params.
        self.grid(mpl.rcParams['axes.grid'])

        # Set the default limits (so that the "native" ticklabels will be
        # correct if they're turned back on)...
        Axes.set_xlim(self, -self.horizon, self.horizon)
        Axes.set_ylim(self, -np.pi / 2.0, np.pi / 2.0)

        # Set up the azimuth ticks.
        self._polar.set_theta_offset(np.radians(self.rotation + 90))
        self._polar.set_theta_direction(-1)
        self._polar.grid(False)
        self._polar.set_rticks([])
    cla.__doc__ = Axes.cla.__doc__

    @property
    def _polar(self):
        """The "hidden" polar axis used for azimuth labels."""
        # This will be called inside LambertAxes.__init__ as well as every
        # time the axis is cleared, so we need the try/except to avoid having
        # multiple hidden axes when `cla` is _manually_ called.
        try:
            return self._hidden_polar_axes
        except AttributeError:
            fig = self.get_figure()
            self._hidden_polar_axes = fig.add_axes(self.get_position(True), 
                                        frameon=False, projection='polar')
            return self._hidden_polar_axes

    def set_azimuth_ticks(self, angles, labels=None, frac=None, **kwargs):
        """
        Sets the azimuthal tick locations (Note: tick lines are not currently
        drawn or supported.).  

        Parameters
        ----------
            angles : A sequence of floats specifying the tick locations in 
                degrees.
            labels : A sequence of strings to use as a label at each location.
                Defaults to a formatted version of the specified angles.
            frac : The radial location of the tick labels. 1.0 is the along the
                edge, 1.1 would be outside, and 0.9 would be inside.
            Additional parameters are text properties for the labels.
        """
        return self._polar.set_thetagrids(angles, labels, frac, **kwargs)

    def get_azimuth_ticks(self, minor=False):
        return self._polar.get_xticks(minor)

    def set_azimuth_ticklabels(self, labels, fontdict=None, **kwargs):
        """
        Sets the labels for the azimuthal ticks.

        Parameters
        ----------
            labels : A sequence of strings
            Additional parameters are text properties for the labels.
        """
        return self._polar.set_xticklabels(labels, fontdict, **kwargs)

    def get_azimuth_ticklabels(self, minor=False):
        """Get the azimuth tick labels as a list of Text artists."""
        return self._polar.get_xticklabels(minor)

    def plane(self, strike, dip, *args, **kwargs):
        """
        Plot lines representing planes on the axes. Additional arguments and
        keyword arguments are passed on to `ax.plot`.

        `strike` and `dip` may be sequences or single values.
    
        Parameters
        ----------
            strike, dip : The strike and dip of the plane(s) in degrees. The 
                dip direction is defined by the strike following the 
                "right-hand rule". 
            segments : The number of segments to use for the line. Defaults
                to 100 segments.
            Additional parameters are passed on to `plot`.

        Returns
        -------
            A sequence of Line2D artists representing the lines specified by
            `strike` and `dip`.
        """
        segments = kwargs.pop('segments', 100)
        lon, lat = stereonet_math.plane(strike, dip, segments)
        return self.plot(lon, lat, *args, **kwargs)

    def pole(self, strike, dip, *args, **kwargs):
        """
        Plot points representing poles to planes on the axes. Additional
        arguments and keyword arguments are passed on to `ax.plot`.

        `strike` and `dip` may be sequences or single values.

        Parameters
        ----------
            strike, dip : The strike and dip of the plane(s) in degrees. The 
                dip direction is defined by the strike following the 
                "right-hand rule". 
            Additional parameters are passed on to `plot`.

        Returns
        -------
            A sequence of Line2D artists representing the point(s) specified by
            `strike` and `dip`.
        """
        lon, lat = stereonet_math.pole(strike, dip)
        args, kwargs = self._point_plot_defaults(args, kwargs)
        return self.plot(lon, lat, *args, **kwargs)

    def rake(self, strike, dip, rake_angle, *args, **kwargs):
        """
        Plot points representing lineations along planes on the axes.
        Additional arguments and keyword arguments are passed on to `plot`.

        `strike`, `dip`, and `rake_angle` may be sequences or single values.

        Parameters
        ----------
            strike, dip : The strike and dip of the plane(s) in degrees. The 
                dip direction is defined by the strike following the 
                "right-hand rule". 
            rake_angle : The angle of the lineation(s) on the plane(s) measured 
                in degrees downward from horizontal. Zero degrees corresponds to 
                the "right hand" direction indicated by the strike, while
                negative angles are measured downward from the opposite strike
                direction.
            Additional arguments are passed on to `plot`.

        Returns
        -------
            A sequence of Line2D artists representing the point(s) specified by
            `strike` and `dip`.
        """
        lon, lat = stereonet_math.rake(strike, dip, rake_angle)
        args, kwargs = self._point_plot_defaults(args, kwargs)
        return self.plot(lon, lat, *args, **kwargs)

    def line(self, plunge, bearing, *args, **kwargs):
        """
        Plot points representing linear features on the axes. Additional
        arguments and keyword arguments are passed on to `plot`.

        `plunge` and `bearing` may be sequences or single values.
    
        Parameters
        ----------
            plunge, bearing : The plunge and bearing of the line(s) in degrees.
                The plunge is measured in degrees downward from the end of the
                feature specified by the bearing. 
            Additional parameters are passed on to `plot`.

        Returns
        -------
            A sequence of Line2D artists representing the point(s) specified by
            `strike` and `dip`.
        """
        lon, lat = stereonet_math.line(plunge, bearing)
        args, kwargs = self._point_plot_defaults(args, kwargs)
        return self.plot([lon], [lat], *args, **kwargs)

    def _point_plot_defaults(self, args, kwargs):
        """To avoid confusion for new users, this ensures that "scattered" 
        points are plotted by by `plot` instead of points joined by a line.
        
        Parameters
        ----------
            args : A tuple of arguments representing additional parameters to be 
                passed to `self.plot`.
            kwargs : A dict of keyword arguments representing additional 
                parameters to be passed to `self.plot`
        Returns
        -------
            Modified versions of `args` and `kwargs`.
        """
        if args:
            return args, kwargs

        if 'ls' not in kwargs and 'linestyle' not in kwargs:
            kwargs['linestyle'] = 'none'
        if 'marker' not in kwargs:
            kwargs['marker'] = 'o'
        return args, kwargs

    def _contour_helper(self, args, kwargs):
        """Unify defaults and common functionality of density_contour and 
        density_contourf."""
        contour_kwargs = {}
        contour_kwargs['measurement'] = kwargs.pop('measurement', 'poles')
        contour_kwargs['method'] = kwargs.pop('method', 'exponential_kamb')
        contour_kwargs['sigma'] = kwargs.pop('sigma', 3)
        contour_kwargs['gridsize'] = kwargs.pop('gridsize', 100)
        contour_kwargs['num_counters'] = kwargs.pop('num_counters', None)
        lon, lat, totals = contouring.density_grid(*args, **contour_kwargs)
        return lon, lat, totals, kwargs

    def density_contour(self, *args, **kwargs):
        """
        Estimates point density of the given linear orientation measurements 
        (Interpreted as poles, lines, rakes, or "raw" longitudes and latitudes
        based on the `measurement` keyword argument.) and plots contour lines of
        the resulting density distribution.

        Parameters
        ----------
            *args : A variable number of sequences of measurements. 
                By default, this will be expected to be `strike` & `dip`, both
                array-like sequences representing poles to planes.  (Rake
                measurements require three parameters, thus the variable number
                of arguments.) The `measurement` kwarg controls how these 
                arguments are interpreted.
            measurement : string, optional 
                Controls how the input arguments are interpreted. Defaults to
                "poles".  May be one of the following:
                    "poles" : Arguments are assumed to be sequences of strikes
                        and dips of planes. Poles to these planes are used for
                        density contouring.
                    "lines" : Arguments are assumed to be sequences of plunges
                        and bearings of linear features.  
                    "rakes" : Arguments are assumed to be sequences of strikes,
                        dips, and rakes along the plane.
                    "radians" : Arguments are assumed to be "raw" longitudes
                        and latitudes in the underlying projection's coordinate
                        system.
            method : string, optional 
                The method of density estimation to use. Defaults to
                "exponential_kamb". May be one of the following:
                "exponential_kamb" : A modified Kamb method using exponential 
                    smoothing _[1]. Units are in numbers of standard deviations
                    by which the density estimate differs from uniform.
                "linear_kamb" : A modified Kamb method using linear smoothing 
                    _[1]. Units are in numbers of standard deviations by which
                    the density estimate differs from uniform.
                "kamb" : Kamb's method _[2] with no smoothing. Units are in
                    numbers of standard deviations by which the density
                    estimate differs from uniform.
                "schmidt" : The traditional "Schmidt" (a.k.a. 1%) method. Counts
                    points within a counting circle comprising 1% of the total
                    area of the hemisphere. Does not take into account sample
                    size. Units are in points per 1% area.
            sigma : int or float, optional
                The number of standard deviations defining the expected number
                of standard deviations by which a random sample from a uniform
                distribution of points would be expected to vary from being 
                evenly distributed across the hemisphere.  This controls the 
                size of the counting circle, and therefore the degree of 
                smoothing.  Higher sigmas will lead to more smoothing of the
                resulting density distribution. This parameter only applies to 
                Kamb-based methods.  Defaults to 3.
            gridsize : int or 2-item tuple of ints, optional
                The size of the grid that the density is estimated on. If a 
                single int is given, it is interpreted as an NxN grid. If a
                tuple of ints is given it is interpreted as (nrows, ncols). 
                Defaults to 100.
            num_counters : int, optional
                The number of "counting points" (arranged following a golden 
                section spiral) that density is estimated at on the surface of
                hemisphere.  This is then interpolated onto a regular grid in
                lat-long space (see "gridsize" above). Defaults to 1/2 of the
                total number of cells in the regular grid.

            Additional keyword arguments are passed on to matplotlib's `contour`
            function.

        Returns:
        --------
            A matplotlib ContourSet.

        See Also:
        ---------
            mplstereonet.density_grid
            mplstereonet.StereonetAxes.density_contourf
            matplotlib.pyplot.contour
            matplotlib.pyplot.clabel


        Examples:
        ---------
            Plot density contours of poles to the specified planes using a
            modified Kamb method with exponential smoothing [1]_.

            >>> strikes, dips = [120, 315, 86], [22, 85, 31]
            >>> ax.density_contour(strikes, dips)

            Plot density contours of a set of linear orientation measurements.

            >>> plunges, bearings = [-10, 20, -30], [120, 315, 86]
            >>> ax.density_contour(plunges, bearings, measurement='lines')

            Plot density contours of a set of rake measurements.

            >>> strikes, dips, rakes = [120, 315, 86], [22, 85, 31], [-5, 20, 9]
            >>> ax.density_contour(strikes, dips, rakes, measurement='rakes')

            Plot density contours of a set of "raw" longitudes and latitudes.

            >>> lon, lat = np.radians([-40, 30, -85]), np.radians([21, -59, 45])
            >>> ax.density_contour(lon, lat, measurement='radians')


            Plot density contours of poles to planes using a Kamb method [2]_
            with the density estimated on a 10x10 grid (in long-lat space) 

            >>> strikes, dips = [120, 315, 86], [22, 85, 31]
            >>> ax.density_contour(strikes, dips, method='kamb', gridsize=10)

            Plot density contours of poles to planes with contours at [1,2,3]
            standard deviations.

            >>> strikes, dips = [120, 315, 86], [22, 85, 31]
            >>> ax.density_contour(strikes, dips, levels=[1,2,3])

        References
        ----------
        .. [1] Vollmer, 1995. C Program for Automatic Contouring of Spherical
           Orientation Data Using a Modified Kamb Method. Computers &
           Geosciences, Vol. 21, No. 1, pp. 31--49.

        .. [2] Kamb, 1959. Ice Petrofabric Observations from Blue Glacier,
           Washington, in Relation to Theory and Experiment. Journal of
           Geophysical Research, Vol. 64, No. 11, pp. 1891--1909.
        """
        lon, lat, totals, kwargs = self._contour_helper(args, kwargs)
        return self.contour(lon, lat, totals, **kwargs)

    def density_contourf(self, *args, **kwargs):
        """
        Estimates point density of the given linear orientation measurements 
        (Interpreted as poles, lines, rakes, or "raw" longitudes and latitudes
        based on the `measurement` keyword argument.) and plots filled contours
        of the resulting density distribution.

        Parameters
        ----------
            *args : A variable number of sequences of measurements. 
                By default, this will be expected to be `strike` & `dip`, both
                array-like sequences representing poles to planes.  (Rake
                measurements require three parameters, thus the variable number
                of arguments.) The `measurement` kwarg controls how these 
                arguments are interpreted.
            measurement : string, optional 
                Controls how the input arguments are interpreted. Defaults to
                "poles".  May be one of the following:
                    "poles" : Arguments are assumed to be sequences of strikes
                        and dips of planes. Poles to these planes are used for
                        density contouring.
                    "lines" : Arguments are assumed to be sequences of plunges
                        and bearings of linear features.  
                    "rakes" : Arguments are assumed to be sequences of strikes,
                        dips, and rakes along the plane.
                    "radians" : Arguments are assumed to be "raw" longitudes
                        and latitudes in the underlying projection's coordinate
                        system.
            method : string, optional 
                The method of density estimation to use. Defaults to
                "exponential_kamb". May be one of the following:
                "exponential_kamb" : A modified Kamb method using exponential 
                    smoothing _[1]. Units are in numbers of standard deviations
                    by which the density estimate differs from uniform.
                "linear_kamb" : A modified Kamb method using linear smoothing 
                    _[1]. Units are in numbers of standard deviations by which
                    the density estimate differs from uniform.
                "kamb" : Kamb's method _[2] with no smoothing. Units are in
                    numbers of standard deviations by which the density
                    estimate differs from uniform.
                "schmidt" : The traditional "Schmidt" (a.k.a. 1%) method. Counts
                    points within a counting circle comprising 1% of the total
                    area of the hemisphere. Does not take into account sample
                    size. Units are in points per 1% area.
            sigma : int or float, optional
                The number of standard deviations defining the expected number
                of standard deviations by which a random sample from a uniform
                distribution of points would be expected to vary from being 
                evenly distributed across the hemisphere.  This controls the 
                size of the counting circle, and therefore the degree of 
                smoothing.  Higher sigmas will lead to more smoothing of the
                resulting density distribution. This parameter only applies to 
                Kamb-based methods.  Defaults to 3.
            gridsize : int or 2-item tuple of ints, optional
                The size of the grid that the density is estimated on. If a 
                single int is given, it is interpreted as an NxN grid. If a
                tuple of ints is given it is interpreted as (nrows, ncols).
                Defaults to 100.
            num_counters : int, optional
                The number of "counting points" (arranged following a golden 
                section spiral) that density is estimated at on the surface of
                hemisphere.  This is then interpolated onto a regular grid in
                lat-long space (see "gridsize" above). Defaults to 1/2 of the
                total number of cells in the regular grid.

            Additional keyword arguments are passed on to matplotlib's
            `contourf` function.

        Returns:
        --------
            A matplotlib `QuadContourSet`.

        See Also:
        ---------
            `mplstereonet.density_grid`
            `mplstereonet.StereonetAxes.density_contour`
            `matplotlib.pyplot.contourf`
            `matplotlib.pyplot.clabel`


        Examples:
        ---------
            Plot filled density contours of poles to the specified planes using
            a modified Kamb method with exponential smoothing [1]_.

            >>> strikes, dips = [120, 315, 86], [22, 85, 31]
            >>> ax.density_contourf(strikes, dips)

            Plot filled density contours of a set of linear orientation
            measurements.

            >>> plunges, bearings = [-10, 20, -30], [120, 315, 86]
            >>> ax.density_contourf(plunges, bearings, measurement='lines')

            Plot filled density contours of a set of rake measurements.

            >>> strikes, dips, rakes = [120, 315, 86], [22, 85, 31], [-5, 20, 9]
            >>> ax.density_contourf(strikes, dips, rakes, measurement='rakes')

            Plot filled density contours of a set of "raw" longitudes and
            latitudes.

            >>> lon, lat = np.radians([-40, 30, -85]), np.radians([21, -59, 45])
            >>> ax.density_contourf(lon, lat, measurement='radians')


            Plot filled density contours of poles to planes using a Kamb method
            [2]_ with the density estimated on a 10x10 grid (in long-lat space) 

            >>> strikes, dips = [120, 315, 86], [22, 85, 31]
            >>> ax.density_contourf(strikes, dips, method='kamb', gridsize=10)

            Plot filled density contours of poles to planes with contours at
            [1,2,3] standard deviations.

            >>> strikes, dips = [120, 315, 86], [22, 85, 31]
            >>> ax.density_contourf(strikes, dips, levels=[1,2,3])

        References
        ----------
        .. [1] Vollmer, 1995. C Program for Automatic Contouring of Spherical
           Orientation Data Using a Modified Kamb Method. Computers &
           Geosciences, Vol. 21, No. 1, pp. 31--49.

        .. [2] Kamb, 1959. Ice Petrofabric Observations from Blue Glacier,
           Washington, in Relation to Theory and Experiment. Journal of
           Geophysical Research, Vol. 64, No. 11, pp. 1891--1909.
        """
        lon, lat, totals, kwargs = self._contour_helper(args, kwargs)
        return self.contourf(lon, lat, totals, **kwargs)


class EqualAngleAxes(StereonetAxes):
    """An axes representing a lower-hemisphere "Wulff" (a.k.a. equal angle) 
    projection."""
    _base_transform = stereonet_transforms.StereographicTransform
    name = 'equal_angle_stereonet'

class EqualAreaAxes(StereonetAxes):
    """An axes representing a lower-hemisphere "Schmitt" (a.k.a. equal area) 
    projection."""
    name = 'equal_area_stereonet'

register_projection(StereonetAxes)
register_projection(EqualAreaAxes)
register_projection(EqualAngleAxes)




import numpy as np
import re
import stereonet_math as smath

def parse_strike_dip(strike, dip):
    """
    Parses strings of strike and dip and returns strike and dip measurements 
    following the right-hand-rule.

    Dip directions are parsed, and if the measurement does not follow the 
    right-hand-rule, the opposite end of the strike measurement is returned.

    Accepts either quadrant-formatted or azimuth-formatted strikes.

    For example, this would convert a strike of "N30E" and a dip of "45NW" to
    a strike of 210 and a dip of 45.

    Parameters:
    -----------
        strike : A string representing a strike measurement. May be in azimuth 
            or quadrant format. 
        dip : A string representing the dip angle and direction of a plane.

    Returns:
    --------
        azi : A float representing the azimuth of the strike of the plane with
            dip direction indicated following the right-hand-rule.
        dip : A float representing the dip of the plane.
    """
    strike = parse_azimuth(strike)
    dip, direction = split_trailing_letters(dip)

    if direction is not None:
        expected_direc = strike + 90
        if opposite_end(expected_direc, direction):
            strike += 180

    if strike > 360:
        strike -= 360

    return strike, dip

def parse_rake(strike, dip, rake):
    """
    Parses strings of strike, dip, and rake and returns a strike, dip, and rake
    measurement following the right-hand-rule, with the "end" of the strike
    that the rake is measured from indicated by the sign of the rake (positive
    rakes correspond to the strike direction, negative rakes correspond to the
    opposite end).

    Accepts either quadrant-formatted or azimuth-formatted strikes.

    For example, this would convert a strike of "N30E", dip of "45NW", with a 
    rake of "10NE" to a strike of 210, dip of 45, and rake of 170.

    Rake angles returned by this function will always be between 0 and 180

    If no directions are specified, the measuriement is assumed to follow the 
    usual right-hand-rule convention.

    Parameters:
    -----------
        strike : A string representing a strike measurement. May be in azimuth 
            or quadrant format. 
        dip : A string representing the dip angle and direction of a plane.
        rake : A string representing the rake angle and direction that the rake
            is measured from.

    Returns:
    --------
        strike, dip, rake : Floating point measurements of strike, dip, and rake
            following the conventions outlined above.
    """
    strike, dip = parse_strike_dip(strike, dip)
    rake, direction = split_trailing_letters(rake)

    if direction is not None:
        if opposite_end(strike, direction):
            rake = -rake

    if rake < 0:
        rake += 180
    elif rake > 180:
        rake -= 180

    return strike, dip, rake

def parse_plunge_bearing(plunge, bearing):
    """
    Parses strings of plunge and bearing and returns a consistent plunge and
    bearing measurement as floats. Plunge angles returned by this function will
    always be between 0 and 90.

    If no direction letter(s) is present, the plunge is assumed to be measured
    from the end specified by the bearing. If a direction letter(s) is present,
    the bearing will be switched to the opposite (180 degrees) end if the
    specified direction corresponds to the opposite end specified by the
    bearing.

    Parameters:
    -----------
        plunge : A string representing a plunge measurement.
        bearing : A string representing a bearing measurement. May be in azimuth
            or quadrant format.

    Examples:
    ---------

        >>> parse_plunge_bearing("30NW", 160)
        ... (30, 340)

    """
    bearing = parse_azimuth(bearing)
    plunge, direction = split_trailing_letters(plunge)

    if direction is not None:
        if opposite_end(bearing, direction):
            bearing +=180

    if plunge < 0:
        bearing += 180
        plunge = -plunge
    if plunge > 90:
        bearing += 180
        plunge = 180 - plunge

    if bearing > 360:
        bearing -= 360

    return plunge, bearing

def opposite_end(azimuth, quadrant):
    direc = quadrantletter_to_azimuth(quadrant)
    dot = np.dot(_azimuth2vec(direc), _azimuth2vec(azimuth))
    return dot < 0

def split_trailing_letters(item):
    # Match things like 45NW, 45E, 45WNW, etc
    letters = re.search(r'[NESWnesw]', item)
    if letters is None:
        return float(item), None
    else:
        measurement = item[:letters.start()]
        direction = item[letters.start():]
        return float(measurement), direction

def circmean(azimuths):
    azimuths = np.radians(azimuths)
    x = np.cos(azimuths)
    y = np.sin(azimuths)
    return np.degrees(np.arctan2(y.mean(), x.mean()))

def quadrantletter_to_azimuth(letters):
    letters = letters.strip()
    azimuth = {'N':0, 'S':180, 'E':90, 'W':270}

    azi = azimuth[letters[-1]]
    for letter in letters[1::-1]:
        azi = circmean([azi, azimuth[letter]])
    return azi

def _azimuth2vec(azi):
    azi = np.radians(azi)
    return np.cos(azi), np.sin(azi)

def parse_azimuth(azimuth):
    """
    Parses an azimuth measurement in azimuth or quadrant format.

    Parameters:
    -----------
        azimuth : A string or number representing an azimuth

    Returns:
    --------
        azi : A float in the range 0-360

    See Also:
    ---------
        `parse_quadrant_measurement`
        `parse_strike_dip`
        `parse_plunge_bearing`
    """
    try:
        azimuth = float(azimuth)
    except ValueError:
        if not azimuth[0].isalpha():
            raise ValueError('Ambiguous azimuth: {}'.format(azimuth))
        azimuth = parse_quadrant_measurement(azimuth)
    return azimuth

def parse_quadrant_measurement(quad_azimuth):
    """
    Parses a quadrant measurement of the form "AxxB", where A and B are cardinal
    directions and xx is an angle measured relative to those directions.

    In other words, it converts a measurement such as E30N into an azimuth of
    60 degrees, or W10S into an azimuth of 260 degrees.

    For ambiguous quadrant measurements such as "N30S", a ValueError is raised.

    Parameters:
    -----------
        quad_azimuth : A string representing a strike measurement in quadrant 
            form.

    Returns:
    --------
        azi : An azimuth in degrees clockwise from north.

    See Also:
    ---------
        `parse_azimuth`
    """
    def rotation_direction(first, second):
        return np.cross(_azimuth2vec(first), _azimuth2vec(second))

    # Parse measurement
    quad_azimuth = quad_azimuth.strip()
    try:
        first_dir = quadrantletter_to_azimuth(quad_azimuth[0].upper())
        sec_dir = quadrantletter_to_azimuth(quad_azimuth[-1].upper())
    except KeyError:
        raise ValueError('{} is not a valid azimuth'.format(quad_azimuth))

    angle = float(quad_azimuth[1:-1])

    # Convert quadrant measurement into an azimuth
    direc = rotation_direction(first_dir, sec_dir)
    azi = first_dir + direc * angle

    # Catch ambiguous measurements such as N10S and raise an error
    if abs(direc) < 0.9:
        raise ValueError('{} is not a valid azimuth'.format(quad_azimuth))
        
    # Ensure that 0 <= azi <= 360
    if azi < 0:
        azi += 360
    elif azi > 360:
        azi -= 360

    return azi




import numpy as np
import stereonet_math
import matplotlib.delaunay as delaunay

def _points_on_hemisphere(num):
    """
    Generates `num` points on a hemisphere using a golden section spiral.
    """
    num = num * 2
    inc = np.pi * (3 - np.sqrt(5))
    off = 2.0 / num
    k = np.arange(num)
    y = k * off - 1 + off / 2.0
    r = np.sqrt(1 - y**2)
    phi = k * inc
    x = np.cos(phi) * r
    z = np.sin(phi) * r
    hemisphere = x > 0
    return x[hemisphere], y[hemisphere], z[hemisphere]

def _count_points(lons, lats, func, sigma, num=10000):
    """This function actually calculates the point density of the input ("lons"
    and "lats") points at a series of "counter stations". Creates "num" counter
    stations in a regular arrangement around a hemisphere, calculate the
    distance to all input points at each counter station, and then calculate
    the density using "func"."""
    # Basically, we can't model this as a convolution as we're not in cartesian
    # space, so we have to iterate throught the counters.
    xyz_counters = np.vstack(_points_on_hemisphere(num)).T
    xyz_points = stereonet_math.sph2cart(lons, lats)
    xyz_points = np.vstack(xyz_points).T

    totals = np.zeros(xyz_counters.shape[0], dtype=np.float)
    for i, xyz in enumerate(xyz_counters):
        cos_dist = np.abs(np.dot(xyz, xyz_points.T))
        totals[i] = func(cos_dist, sigma)

    totals[totals < 0] = 0
    counter_lon, counter_lat = stereonet_math.cart2sph(*xyz_counters.T)
    return counter_lon, counter_lat, totals

def _grid_data(lons, lats, z, gridsize=(100,100)):
    """Interpolate density data onto a regular grid between -pi/2 to pi/2 in 
    x and y."""
    bound = np.pi / 2.0 + 0.1 # We need to go a bit beyond the bounds...
    nrows, ncols = gridsize
    xmin, xmax, ymin, ymax = -bound, bound, -bound, bound

    tri = delaunay.Triangulation(lons, lats)
    interp = delaunay.LinearInterpolator(tri, z, default_value=0)

    slices = np.s_[xmin : xmax : ncols * 1j, ymin : ymax : nrows * 1j]
    zi = interp[slices]

    xi, yi = np.ogrid[slices]
    return xi.ravel(), yi.ravel(), zi

def density_grid(*args, **kwargs):
    """
    Estimates point density of the given linear orientation measurements
    (Interpreted as poles, lines, rakes, or "raw" longitudes and latitudes
    based on the `measurement` keyword argument.). Returns a regular (in
    lat-long space) grid of density estimates over a hemispherical surface.

    Parameters
    ----------

        *args : A variable number of sequences of measurements. By default, this
            will be expected to be `strike` & `dip`, both array-like sequences
            representing poles to planes.  (Rake measurements require three
            parameters, thus the variable number of arguments.) The
            `measurement` kwarg controls how these arguments are interpreted.
        measurement : string, optional 
            Controls how the input arguments are interpreted. Defaults to
            "poles".  
            May be one of the following:
                "poles" : Arguments are assumed to be sequences of strikes and 
                    dips of planes. Poles to these planes are used for density
                    contouring.
                "lines" : Arguments are assumed to be sequences of plunges and
                    bearings of linear features.  
                "rakes" : Arguments are assumed to be sequences of strikes,
                    dips, and rakes along the plane.
                "radians" : Arguments are assumed to be "raw" longitudes and
                    latitudes in the underlying projection's coordinate system.
        method : string, optional 
            The method of density estimation to use. Defaults to
            "exponential_kamb". 
            May be one of the following:
                "exponential_kamb" : A modified Kamb method using exponential 
                    smoothing _[1]. Units are in numbers of standard deviations
                    by which the density estimate differs from uniform.
                "linear_kamb" : A modified Kamb method using linear smoothing 
                    _[1]. Units are in numbers of standard deviations by which
                    the density estimate differs from uniform.
                "kamb" : Kamb's method _[2] with no smoothing. Units are in
                    numbers of standard deviations by which the density
                    estimate differs from uniform.
                "schmidt" : The traditional "Schmidt" (a.k.a. 1%) method. Counts
                    points within a counting circle comprising 1% of the total
                    area of the hemisphere. Does not take into account sample
                    size. Units are in points per 1% area.
        sigma : int or float, optional
            The number of standard deviations defining the expected number of
            standard deviations by which a random sample from a uniform
            distribution of points would be expected to vary from being evenly
            distributed across the hemisphere.  This controls the size of the
            counting circle, and therefore the degree of smoothing.  Higher
            sigmas will lead to more smoothing of the resulting density
            distribution. This parameter only applies to Kamb-based methods.
            Defaults to 3.
        gridsize : int or 2-item tuple of ints, optional
            The size of the grid that the density is estimated on. If a single
            int is given, it is interpreted as an NxN grid. If a tuple of ints
            is given it is interpreted as (nrows, ncols).  Defaults to 100.
        num_counters : int, optional
            The number of "counting points" (arranged following a golden
            section spiral) that density is estimated at on the surface of
            hemisphere.  This is then interpolated onto a regular grid in
            lat-long space (see "gridsize" above). Defaults to 1/2 of the total
            number of cells in the regular grid.

    Returns:
    --------
        xi, yi, zi : The longitude, latitude and density values of the regularly
            gridded density estimates. Longitude and latitude are in radians.
        
    See Also:
    ---------
        `mplstereonet.StereonetAxes.density_contourf`
        `mplstereonet.StereonetAxes.density_contour`

    References
    ----------
    .. [1] Vollmer, 1995. C Program for Automatic Contouring of Spherical
       Orientation Data Using a Modified Kamb Method. Computers &
       Geosciences, Vol. 21, No. 1, pp. 31--49.

    .. [2] Kamb, 1959. Ice Petrofabric Observations from Blue Glacier,
       Washington, in Relation to Theory and Experiment. Journal of
       Geophysical Research, Vol. 64, No. 11, pp. 1891--1909.
    """
    def do_nothing(x, y):
        return x, y
    measurement = kwargs.get('measurement', 'poles')
    gridsize = kwargs.get('gridsize', 100)
    try:
        gridsize = int(gridsize)
        gridsize = (gridsize, gridsize)
    except TypeError:
        pass

    num_counters = kwargs.get('num_counters', None)
    if num_counters is None:
        num_counters = np.product(gridsize) / 2

    func = {'poles':stereonet_math.pole,
            'lines':stereonet_math.line,
            'rakes':stereonet_math.rake,
            'radians':do_nothing}[measurement]
    lon, lat = func(*args)

    method = kwargs.get('method', 'exponential_kamb')
    sigma = kwargs.get('sigma', 3)
    func = {'linear_kamb':_linear_inverse_kamb,
            'square_kamb':_square_inverse_kamb,
            'schmidt':_schmidt_count,
            'kamb':_kamb_count,
            'exponential_kamb':_exponential_kamb,
            }[method]
    counter_lon, counter_lat, totals = _count_points(lon, lat, func, sigma,
                                                     num_counters)
    xi, yi, zi = _grid_data(counter_lon, counter_lat, totals, gridsize)
    return xi, yi, zi

def _exponential_kamb(cos_dist, sigma=3):
    """Kernel function from Vollmer for exponential smoothing."""
    n = float(cos_dist.size)
    f = 2 * (1.0 + n / sigma**2)
    count = np.exp(f * (cos_dist - 1)).sum()
    units = np.sqrt(n * (f/2.0 - 1) / f**2) 
    return (count - 0.5) / units 

def _linear_inverse_kamb(cos_dist, sigma=3):
    """Kernel function from Vollmer for linear smoothing."""
    n = float(cos_dist.size)
    radius = _kamb_radius(n, sigma)
    f = 2 / (1 - radius)
    cos_dist = cos_dist[cos_dist >= radius]
    count = (f * (cos_dist - radius)).sum()
    return (count - 0.5) / _kamb_units(n, radius) 

def _square_inverse_kamb(cos_dist, sigma=3):
    """Kernel function from Vollemer for inverse square smoothing."""
    n = float(cos_dist.size)
    radius = _kamb_radius(n, sigma)
    f = 3 / (1 - radius)**2
    cos_dist = cos_dist[cos_dist >= radius]
    count = (f * (cos_dist - radius)**2).sum()
    return (count - 0.5) / _kamb_units(n, radius)

def _kamb_radius(n, sigma):
    """Radius of kernel for Kamb-style smoothing."""
    a = sigma**2 / (float(n) + sigma**2)
    return (1 - a) 

def _kamb_units(n, radius):
    """Normalization function for Kamb-style counting."""
    return np.sqrt(n * radius * (1 - radius))

def _kamb_count(cos_dist, sigma=3):
    """Original Kamb kernel function (raw count within radius)."""
    n = float(cos_dist.size)
    dist = _kamb_radius(n, sigma)
    count = (cos_dist >= dist).sum()
    return (count - 0.5) / _kamb_units(n, dist)

def _schmidt_count(cos_dist, sigma=None):
    """Schmidt (a.k.a. 1%) counting kernel function."""
    radius = 0.01
    count = ((1 - cos_dist) <= radius).sum()
    return (count - 0.5) / (cos_dist.size * radius)



"""
Utilities to convert between strike/dip, etc and points/lines in lat, long
space.

A stereonet in <long,lat> coordinates:
              <0,90>
               ***
            *       *
   <-90,0> *         *<90,0>
           *         *
            *       *
               ***
             <0,-90>                  

If strike=0, plotting lines, rakes, planes or poles to planes is simple.  For a
plane, it's a line of constant longitude at long=90-dip.  For a line, it's a
point at long=0,lat=90-dip.  For a rake, it's a point at long=90-dip,
lat=90-rake.  These points can then be rotated to the proper strike. (A
rotation matrix around the X-axis is much simpler than the trig otherwise
necessary!)

All of these assume that strikes and dips follow the "right-hand-rule". 
In other words, if we're facing in the direction given for the strike, the plane
dips to our right.
"""
import numpy as np

def sph2cart(lon, lat):
    """
    Converts a longitude and latitude (or sequence of lons and lats) given in 
    _radians_ to cartesian coordinates, `x`, `y`, `z`, where x=0, y=0, z=0 is
    the center of the globe.

    Parameters
    ----------
        lon : Longitude in radians
        lat : Latitude in radians

    Returns
    -------
        `x`, `y`, `z` : Arrays of cartesian coordinates
    """
    x = np.cos(lat)*np.cos(lon)
    y = np.cos(lat)*np.sin(lon)
    z = np.sin(lat)
    return x, y, z

def cart2sph(x, y, z):
    """
    Converts cartesian coordinates `x`, `y`, `z` into a longitude and latitude.
    x=0, y=0, z=0 is assumed to correspond to the center of the globe.
    Returns lon and lat in radians.

    Parameters
    ----------
        `x`, `y`, `z` : Arrays of cartesian coordinates

    Returns
    -------
        lon : Longitude in radians
        lat : Latitude in radians
    """
    r = np.sqrt(x**2 + y**2 + z**2)
    lat = np.arcsin(z/r)
    lon = np.arctan2(y, x)
    return lon, lat

def _rotate(lon, lat, theta):
    """
    Rotate "lon", "lat" coords (in degrees) about the X-axis by "theta" degrees. 
    This effectively simulates rotating a physical stereonet.
    Returns rotated lon, lat coords in radians.
    """
    # Convert input to numpy arrays in radians
    lon, lat = np.atleast_1d(lon, lat)
    lon, lat = map(np.radians, [lon, lat])
    theta = np.radians(theta)

    # Convert to cartesian coords for the rotation
    x, y, z = sph2cart(lon, lat)

    # This is just a rotation matrix for a rotation about the X-axis
    X = x
    Y = y*np.cos(theta) + z*np.sin(theta)
    Z = -y*np.sin(theta) + z*np.cos(theta)

    # Now convert back to spherical coords (longitude and latitude, ignore R)
    lon, lat = cart2sph(X,Y,Z) 
    return lon, lat # in radians!

def plane(strike, dip, segments=100):
    """
    Calculates the longitude and latitude of `segments` points along the 
    stereonet projection of each plane with a given `strike` and `dip` in 
    degrees.
    
    `strike` and `dip` may be sequences or single values.
    
    Parameters
    ----------
        strike : The strike of the plane(s) in degrees, with dip direction 
            indicated by the azimuth (e.g. 315 vs. 135) specified following the
            "right hand rule". 
        dip : The dip of the plane(s) in degrees.         
        segments : The number of points in the returned `lon` and `lat` arrays.
            Defaults to 100 segments.
    
    Returns:
        lon, lat : num_segments x num_strikes arrays of longitude and latitude
            in radians.
    """
    strikes, dips = np.atleast_1d(strike, dip)
    lons = np.zeros((segments, strikes.size), dtype=np.float)
    lats = lons.copy()
    for i, (strike, dip) in enumerate(zip(strikes, dips)):
        # We just plot a line of constant longitude and rotate it by the strike.
        dip = 90 - dip
        lon = dip * np.ones(segments)
        lat = np.linspace(-90, 90, segments)
        lon, lat = _rotate(lon, lat, strike)
        lons[:,i] = lon
        lats[:,i] = lat
    return lons, lats

def pole(strike, dip):
    """
    Calculates the longitude and latitude of the pole(s) to the plane(s) 
    specified by `strike` and `dip`, given in degrees.

    `strike` and `dip` may be sequences or single values.
    
    Parameters
    ----------
        strike : The strike of the plane(s) in degrees, with dip direction 
            indicated by the azimuth (e.g. 315 vs. 135) specified following the
            "right hand rule". 
        dip : The dip of the plane(s) in degrees.     

    Returns:
        lon, lat : Arrays of longitude and latitude in radians.
    """
    strike, dip = np.atleast_1d(strike, dip)
    # Plot the approriate point for a strike of 0 and rotate it
    lon, lat = -dip, 0.0
    lon, lat = _rotate(lon, lat, strike)
    return lon, lat

def rake(strike, dip, rake_angle):
    """
    Calculates the longitude and latitude of the linear feature(s) specified by
    `strike`, `dip`, and `rake_angle`.

    `strike`, `dip`, and `rake_angle` may be sequences or single values.
    
    Parameters
    ----------
        strike : The strike of the plane(s) in degrees, with dip direction 
            indicated by the azimuth (e.g. 315 vs. 135) specified following the
            "right hand rule". 
        dip : The dip of the plane(s) in degrees. 
        rake_angle : The angle of the lineation on the plane measured in degrees
            downward from horizontal. Zero degrees corresponds to the "right-
            hand" direction indicated by the strike, while 180 degrees or a 
            negative angle corresponds to the opposite direction.
    
    Returns:
        lon, lat : Arrays of longitude and latitude in radians.
    """
    strike, dip, rake_angle = np.atleast_1d(strike, dip, rake_angle)
    # Plot the approriate point for a strike of 0 and rotate it
    dip = 90 - dip
    lon = dip

    rake_angle = rake_angle.copy()
    rake_angle[rake_angle < 0] += 180
    lat = 90 - rake_angle

    lon, lat = _rotate(lon, lat, strike)
    return lon, lat

def line(plunge, bearing):
    """
    Calculates the longitude and latitude of the linear feature(s) specified by
    `plunge` and `bearing`.

    `plunge` and `bearing` may be sequences or single values.
    
    Parameters
    ----------
        plunge : The plunge of the line(s) in degrees. The plunge is measured 
            in degrees downward from the end of the feature specified by the 
            bearing.         
        bearing : The bearing (azimuth) of the line(s) in degrees.

    Returns:
        lon, lat : Arrays of longitude and latitude in radians.
    """
    plunge, bearing = np.atleast_1d(plunge, bearing)
    # Plot the approriate point for a bearing of 0 and rotate it
    lat = 90 - plunge
    lon = 0
    lon, lat = _rotate(lon, lat, bearing)
    return lon, lat

def plunge_bearing2pole(plunge, bearing):
    """
    Converts the given `plunge` and `bearing` in degrees to a strike and dip
    of the plane whose pole would be parallel to the line specified. (i.e. The
    pole to the plane returned would plot at the same point as the specified
    plunge and bearing.)

    `plunge` and `bearing` may be sequences or single values.
    
    Parameters
    ----------
        plunge : The plunge of the line(s) in degrees. The plunge is measured 
            in degrees downward from the end of the feature specified by the 
            bearing.         
        bearing : The bearing (azimuth) of the line(s) in degrees.

    Returns:
    --------
        strike, dip : Arrays of strikes and dips in degrees following the 
            right-hand-rule.
    """
    plunge, bearing = np.atleast_1d(plunge, bearing)
    strike = bearing + 90
    dip = 90 - plunge
    strike[strike >= 360] -= 360
    return strike, dip

def mean_vector(lons, lats):
    """
    Returns the resultant vector from a series of longitudes and latitudes

    Parameters:
    -----------
        lons : A sequence of longitudes (in radians)
        lats : A sequence of latitudes (in radians)

    Returns:
        mean_vec : A tuple of (lon, lat) in radians
        r_value : The magnitude of the resultant vector (between 0 and 1)
            This represents the degree of clustering in the data.
    """
    xyz = sph2cart(lons, lats)
    xyz = np.vstack(xyz).T
    mean_vec = xyz.mean(axis=0)
    r_value = np.linalg.norm(mean_vec) / xyz.shape[0]
    mean_vec = cart2sph(*mean_vec)
    return mean_vec, r_value

def geographic2pole(lon, lat):
    """
    Converts a longitude and latitude (from a stereonet) into the strike and dip
    of the plane whose pole lies at the given longitude(s) and latitude(s).

    Parameters:
    -----------
        lon : A sequence of longitudes (or a single longitude) in radians
        lat : A sequence of latitudes (or a single latitude) in radians

    Returns:
    --------
        strike : A sequence of strikes in degrees
        dip : A sequence of dips in degrees
    """
    plunge, bearing = geographic2plunge_bearing(lon, lat)
    strike = bearing + 90 
    strike[strike >= 360] -= 360
    dip = 90 - plunge
    return strike, dip

def geographic2plunge_bearing(lon, lat):
    """
    Converts longitude and latitude in stereonet coordinates into a 
    plunge/bearing.

    Parameters:
    -----------
        lon, lat : Sequences of (or single) longitudes and latitudes in radians
            as measured from a lower-hemisphere stereonet

    Returns:
    --------
        plunge : The plunge of the vector in degrees downward from horizontal.
        bearing : The bearing of the vector in degrees clockwise from north.
    """
    lon, lat = np.atleast_1d(lon, lat)
    x, y, z = sph2cart(lon, lat)

    # Bearing will be in the y-z plane...
    bearing = np.arctan2(z, y)

    # Plunge is the angle between the line and the y-z plane
    r = np.sqrt(x*x + y*y + z*z)
    r[r == 0] = 1e-15
    plunge = np.arcsin(x / r)

    # Convert back to azimuths in degrees..
    plunge, bearing = np.degrees(plunge), np.degrees(bearing)
    bearing = 90 - bearing
    bearing[bearing < 0] += 360

    # If the plunge angle is upwards, get the opposite end of the line
    upwards = plunge < 0
    plunge[upwards] *= -1
    bearing[upwards] -= 180
    bearing[upwards & (bearing < 0)] += 360

    return plunge, bearing

def plane_intersection(strike1, dip1, strike2, dip2):
    """
    Finds the intersection of two planes. Returns a plunge/bearing of the linear
    intersection of the two planes.

    Also accepts sequences of strike1s, dip1s, strike2s, dip2s.

    Parameters:
    -----------
        strike1, dip1 : The strike and dip (in degrees, following the 
            right-hand-rule) of the first plane(s).
        strike2, dip2 : The strike and dip (in degrees, following the 
            right-hand-rule) of the second plane(s).

    Returns:
    --------
        plunge, bearing : The plunge and bearing(s) (in degrees) of the line 
            representing the intersection of the two planes.
    """
    norm1 = sph2cart(*pole(strike1, dip1))
    norm2 = sph2cart(*pole(strike2, dip2))
    norm1, norm2 = np.array(norm1), np.array(norm2)
    lon, lat = cart2sph(*np.cross(norm1, norm2, axis=0))
    return geographic2plunge_bearing(lon, lat)

def project_onto_plane(strike, dip, plunge, bearing):
    """
    Projects a linear feature(s) onto the surface of a plane. Returns a rake 
    angle(s) along the plane.

    This is also useful for finding the rake angle of a feature that already
    intersects the plane in question.

    Parameters:
    -----------
        strike, dip : The strike and dip (in degrees, following the 
            right-hand-rule) or a sequence of strikes and dips of the plane(s).
        plunge, bearing : The plunge and bearing (in degrees) or a sequence of
            plunges and bearings of the linear feature(s) to be projected onto
            the plane.

    Returns:
    --------
        rake : A sequence of rake angles measured downwards from horizontal in 
            degrees.  Zero degrees corresponds to the "right- hand" direction
            indicated by the strike, while a negative angle corresponds to the
            opposite direction. Rakes returned by this function will always be
            between -90 and 90 (inclusive).
    """
    # Project the line onto the plane
    norm = sph2cart(*pole(strike, dip))
    feature = sph2cart(*line(plunge, bearing))
    norm, feature = np.array(norm), np.array(feature)
    perp = np.cross(norm, feature, axis=0)
    on_plane = np.cross(perp, norm, axis=0)
    on_plane /= np.sqrt(np.sum(on_plane**2, axis=0))

    # Calculate the angle between the projected feature and horizontal
    # This is just a dot product, but we need to work with multiple measurements
    # at once, so einsum is quicker than apply_along_axis.
    strike_vec = sph2cart(*line(0, strike))
    dot = np.einsum('ij,ij->j', on_plane, strike_vec)
    rake = np.degrees(np.arccos(dot))

    # Convert rakes over 90 to negative rakes...
    rake[rake > 90] -= 180
    rake[rake < -90] += 180
    return rake

def azimuth2rake(strike, dip, azimuth):
    """
    Projects an azimuth of a linear feature onto a plane as a rake angle.

    Parameters:
    -----------
        strike, dip : The strike and dip of the plane in degrees following the
            right-hand-rule.
        azimuth : The azimuth of the linear feature in degrees clockwise from
            north (i.e. a 0-360 azimuth).

    Returns:
    --------
        rake : A rake angle in degrees measured downwards from horizontal. 
            Negative values correspond to the opposite end of the strike.
    """
    plunge, bearing = plane_intersection(strike, dip, azimuth, 90)
    rake = project_onto_plane(strike, dip, plunge, bearing)
    return rake

def xyz2stereonet(x, y, z):
    """
    Converts x, y, z in _world_ cartesian coordinates into lower-hemisphere
    stereonet coordinates.

    Parameters
    ----------
        x, y, z : Sequences of world coordinates

    Returns:
    --------
        lon, lat : Sequences of longitudes and latitudes (in radians)
    """
    x, y, z = np.atleast_1d(x, y, z)
    return cart2sph(-z, x, y)

def stereonet2xyz(lon, lat):
    """
    Converts a sequence of longitudes and latitudes from a lower-hemisphere
    stereonet into _world_ x,y,z coordinates.

    Parameters
    ----------
        lon, lat : Sequences of longitudes and latitudes (in radians) from a
            lower-hemisphere stereonet

    Returns:
    --------
        x, y, z : The world x,y,z components of the vectors represented by the 
            lon, lat coordinates on the stereonet.
    """
    lon, lat = np.atleast_1d(lon, lat)
    x, y, z = sph2cart(lon, lat)
    return y, z, -x

def vector2plunge_bearing(x, y, z):
    """
    Converts a vector or series of vectors given as x, y, z in world
    coordinates into plunge/bearings.

    Parameters:
    -----------
        x : The x-component(s) of the normal vector
        y : The y-component(s) of the normal vector
        z : The z-component(s) of the normal vector

    Returns:
    --------
        plunge : The plunge of the vector in degrees downward from horizontal.
        bearing : The bearing of the vector in degrees clockwise from north.
    """
    return geographic2plunge_bearing(*xyz2stereonet(x,y,z))

def vector2pole(x, y, z):
    """
    Converts a vector or series of vectors given as x, y, z in world
    coordinates into the strike/dip of the planes whose normal vectors are 
    parallel to the specified vectors.  (In other words, each xi,yi,zi is 
    treated as a normal vector and this returns the strike/dip of the 
    corresponding plane.)

    Parameters:
    -----------
        x : The x-component(s) of the normal vector
        y : The y-component(s) of the normal vector
        z : The z-component(s) of the normal vector

    Returns:
    --------
        strike : The strike of the plane, in degrees clockwise from north.
            Dip direction is indicated by the "right hand rule".
        dip : The dip of the plane, in degrees downward from horizontal
    """
    return  geographic2pole(*xyz2stereonet(x, y, z))





from setuptools import setup

setup(
    name = 'mplstereonet',
    version = '0.3',
    description = "Stereonets for matplotlib",
    author = 'Joe Kington',
    author_email = 'joferkington@gmail.com',
    license = 'MIT',
    url = 'https://github.com/joferkington/mplstereonet/',
    packages = ['mplstereonet'],
    install_requires = [
        'matplotlib >= 1.1',
        'numpy >= 1.1']
)
